from textwrap import dedent
from infi.exceptools import chain, InfiException
from logging import getLogger

logger = getLogger(__name__)

def parse_docopt_string(docopt_string):
    """returns a 2-tuple (usage, options)"""
    from re import match, DOTALL
    only_usage_pattern = r"""\s+Usage:\s+(?P<usage>.*)\s+"""
    usage_and_options_pattern = r"""\s+Usage:\s+(?P<usage>.*)\s+Options:\s+(?P<options>.*)\s+"""
    usage, options = '', ''
    if match(usage_and_options_pattern, docopt_string, DOTALL):
        usage = match(usage_and_options_pattern, docopt_string, DOTALL).groupdict()['usage']
        options = match(usage_and_options_pattern, docopt_string, DOTALL).groupdict()['options']
    elif match(only_usage_pattern, docopt_string, DOTALL):
        usage = match(only_usage_pattern, docopt_string, DOTALL).groupdict()['usage']
    return usage, options

def build_usage_and_options():
    from infi.projector.plugins import plugin_repository
    usage = ''
    options = ''
    for plugin in plugin_repository.get_all_plugins():
        docopt_string = plugin.get_docopt_string()
        plugin_usage, plugin_options = parse_docopt_string(docopt_string)
        usage = '\n'.join([usage, plugin_usage])
        options = '\n'.join([options, plugin_options])
    return usage, options

DEFAULT_USAGE = dedent("-h | --help\n-v | --version")
DEFAULT_OPTIONS = dedent("""-h            Show this screen.
                          -v            Show version.""")
SCRIPT_NAME = "projector"

def indent_usage(string):
    lines = [line.strip() for line in string.splitlines() if line.strip() != '']
    return '\n'.join(['    {}'.format(line if line.startswith(SCRIPT_NAME) else "{} {}".format(SCRIPT_NAME, line))
                      for line in lines])

def ident_options(string):
    lines = [line.strip() for line in string.splitlines() if line.strip() != '']
    options_tuple = [(item[0].strip(), item[1].strip()) for item in [line.split('  ', 1) for line in lines]]
    max_option_length = max([len(option_tuple[0]) for option_tuple in options_tuple]) + 4
    return '\n'.join(['    {}{}'.format(option_tuple[0].ljust(max_option_length), option_tuple[1])
                      for option_tuple in options_tuple])

def get_version():
    try:
        from infi.projector.__version__ import __version__
        return __version__
    except ImportError: # pragma: no cover
        return '<unknown>'

def get_commandline_doc():
    __version__ = get_version()
    usage, options = build_usage_and_options()
    doc = "{script_name} {version}\n\nUsage:\n{usage}\n\nOptions:\n{options}\n"
    all_usage = '\n'.join([usage, DEFAULT_USAGE])
    all_options = '\n'.join([DEFAULT_OPTIONS, options])
    return doc.format(script_name=SCRIPT_NAME,
                      usage=indent_usage(all_usage),
                      options=ident_options(all_options),
                      version=__version__)

class ParseException(InfiException):
    pass

def parse_configfile(configfile_path):
    from ConfigParser import ConfigParser
    try:
        parser = ConfigParser()
        parser.read(configfile_path)
    except:
        raise chain(InfiException())
    return parser

def parse_configfile_value(value):
    try:
        return eval(value)
    except:
        return value

def merge_commandline_arguments_from_configfile(arguments, configfile_path):
    try:
        configuration = parse_configfile(configfile_path)
        logger.debug("Failed to parse {}".format(configfile_path))
    except ParseException:
        return
    if not configuration.has_section("commandline-arguments"):
        logger.debug("File {} has no commandline-arguments".format(configfile_path))
        return
    for name, value in configuration.items("commandline-arguments"):
        if arguments.get('name') in [None, False, list()]:
            logger.debug("Setting commandline-argument {} to {}".format(name, value))
            arguments[name] = parse_configfile_value(value)

def append_default_arguments_from_configuration_files(arguments):
    from os import curdir, path
    CONFIGURATION_FILES = [path.expanduser(path.join("~", ".projector")),
                           path.join(curdir, ".projector")]
    for configfile_path in CONFIGURATION_FILES:
        merge_commandline_arguments_from_configfile(arguments, configfile_path)

def parse_commandline_arguments(argv):
    from infi.projector.plugins import plugin_repository
    from docopt import docopt
    doc = get_commandline_doc()
    arguments = dict(docopt(doc, argv=argv, version=get_version()))
    if arguments.get('-v'):
        print get_version()
        return
    plugins = {plugin.get_command_name():plugin for plugin in plugin_repository.get_all_plugins()}
    [selected_plugin] = [value for key, value in plugins.items() if arguments.get(key)]
    append_default_arguments_from_configuration_files(arguments)
    selected_plugin.parse_commandline_arguments(arguments)



from infi.projector.helper import assertions
from infi.projector.helper.utils import open_buildout_configfile

#pylint: disable=R0923

def to_dict(iterable):
    from pkg_resources import parse_requirements
    return {item.project_name: item.specs for item in parse_requirements('\n'.join(iterable))}

def specs_to_string(value):
    return ''.join(value[0]) if value else ''

def from_dict(dict_object):
    return set(['{}{}'.format(key, specs_to_string(value)) for key, value in dict_object.items()])

class PackageSetInterface(object): # pragma: no cover
    def get(self):
        raise NotImplementedError()

    def set(self, package_set):
        raise NotImplementedError()

class BasePackageSet(PackageSetInterface, object):
    def __init__(self, section_name, attribute_name):
        super(BasePackageSet, self).__init__()
        self.attribute_name = attribute_name
        self.section_name = section_name

    def get(self):
        with open_buildout_configfile() as buildout_cfg:
            return self.from_value(buildout_cfg.get(self.section_name, self.attribute_name))

    def set(self, package_set):
        with open_buildout_configfile(write_on_exit=True) as buildout_cfg:
            buildout_cfg.set(self.section_name, self.attribute_name, self.to_value(package_set))

    @classmethod
    def from_value(cls, value): # pragma: no cover
        raise NotImplementedError()

    @classmethod
    def to_value(cls, package_set): # pragma: no cover
        raise NotImplementedError()

class RepresentedListSet(BasePackageSet):
    @classmethod
    def from_value(cls, value):
        return set([item.replace(' ', '') for item in set(eval(value))])

    @classmethod
    def to_value(cls, package_set):
        return repr(list([item.replace(' ', '') for item in set(package_set)]))

class MultilineValueSet(BasePackageSet):
    @classmethod
    def from_value(cls, value):
        return set(value.splitlines())

    @classmethod
    def to_value(cls, package_set):
        newline = '\r\n' if assertions.is_windows() else '\n'
        return newline.join(list(set(package_set)))

class EntryPointSet(BasePackageSet):
    @classmethod
    def from_value(cls, value):
        formatted_entrypoints = eval(value)
        return {name.strip():entry_point.strip()
                for name, entry_point in [item.split('=') for item in formatted_entrypoints]}

    @classmethod
    def to_value(cls, package_set):
        return ["{} = {}".format(key, value) for key, value in package_set.items()]

class InstallRequiresPackageSet(RepresentedListSet):
    def __init__(self):
        super(InstallRequiresPackageSet, self).__init__('project', 'install_requires')

class EggsPackageSet(MultilineValueSet):
    def __init__(self):
        super(EggsPackageSet, self).__init__(self.get_section(), "eggs")

    @classmethod
    def get_section(cls):
        with open_buildout_configfile() as buildout:
            sections = [section for section in buildout.sections()
                        if buildout.has_option(section, "recipe") and \
                        buildout.get(section, "recipe") == "infi.recipe.console_scripts"]
            return sections[0]

class ConsoleScriptsSet(EntryPointSet):
    def __init__(self):
        super(ConsoleScriptsSet, self).__init__('project', 'console_scripts')

class PackageDataSet(RepresentedListSet):
    def __init__(self):
        super(PackageDataSet, self).__init__('project', 'package_data')

class VersionSectionSet(PackageSetInterface, object):
    def __init__(self, filepath="versions.cfg", section_name="versions"):
        super(VersionSectionSet, self).__init__()
        self.filepath = filepath
        self.section_name = section_name

    def get(self):
        with open_buildout_configfile(self.filepath) as cfg:
            return self.from_value(cfg)

    def set(self, package_set):
        raise NotImplementedError()

    @classmethod
    def from_value(cls, cfg): # pragma: no cover
        names = set()
        items = set()
        for option in cfg.options("versions"):
            if option in names:
                continue
            items.add("{}=={}".format(option, cfg.get("versions", option)))
        return items

    @classmethod
    def to_value(cls, package_set): # pragma: no cover
        raise NotImplementedError()



__version__ = "0.5.3"
__git_commiter_name__ = "Guy Rozendorn"
__git_commiter_email__ = "guy@rzn.co.il"
__git_branch__ = '(Not currently on any branch)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '97fe79e50a56b386554f47cc4433c4645495b72a'
__git_head_subject__ = 'Finished release v0.5.3.'
__git_head_message__ = '* release/v0.5.3:\n  Freezing dependencies\n  TRIVIAL --push-changes on freeze fix\n  TRIVIAL --push-changes on freeze fix\n  TRIVIAL --push-changes on freeze\n  Unfreezing dependencies\n  empty commit after version v0.5.2'
__git_dirty_diff__ = ''


from contextlib import contextmanager
from infi.projector.plugins import CommandPlugin
from infi.projector.helper import assertions
from infi.projector.helper.utils import open_buildout_configfile, commit_changes_to_buildout
from infi.projector.helper.utils.package_sets import ConsoleScriptsSet
from textwrap import dedent
from logging import getLogger

logger = getLogger(__name__)

USAGE = """
Usage:
    projector console-scripts list
    projector console-scripts add <script-name> <entry-point> [--commit-changes]
    projector console-scripts remove <script-name> [--commit-changes]
"""

class ConsoleScriptsPlugin(CommandPlugin):
    def get_docopt_string(self):
        return USAGE

    def get_command_name(self):
        return 'console-scripts'

    @assertions.requires_repository
    def parse_commandline_arguments(self, arguments):
        methods = [self.list, self.add, self.remove]
        [method] = [method for method in methods
                    if arguments.get(method.__name__)]
        self.arguments = arguments
        method()

    def get_set(self):
        return ConsoleScriptsSet()

    def list(self):
        from pprint import pprint
        pprint(self.get_set().get())

    def remove(self):
        package_set = self.get_set()
        console_scripts = package_set.get()
        console_script = self.arguments.get('<script-name>')
        if console_script in console_scripts.keys():
            console_scripts.pop(console_script)
            package_set.set(console_scripts)
        if self.arguments.get("--commit-changes", False):
            commit_message = "removing {} from console_scripts".format(console_script)
            commit_changes_to_buildout(commit_message)

    def add(self):
        package_set = self.get_set()
        console_scripts = package_set.get()
        script_name = self.arguments.get('<script-name>')
        console_scripts[script_name] = self.arguments.get('<entry-point>')
        package_set.set(console_scripts)
        if self.arguments.get("--commit-changes", False):
            commit_message = "adding {} to console_scripts".format(script_name)
            commit_changes_to_buildout(commit_message)




from contextlib import contextmanager
from logging import getLogger
logger = getLogger(__name__)

BUILDOUT_PARAMETERS = ['-s']

class PrettyExecutionError(Exception):
    # infi.execute.ExecutionError does print stdout and stderr well, and this is a must when running buildout
    def __init__(self, result):
        super(PrettyExecutionError, self).__init__("Execution of %r failed!\nresult=%s\nstdout=%s\nstderr=%s" % (result._command,
                                                                                                                 result.get_returncode(),
                                                                                                                 result.get_stdout(),
                                                                                                                 result.get_stderr()))
        self.result = result

def _chdir_and_log(path):
    from os import chdir
    chdir(path)
    logger.debug("Changed directory to {!r}".format(path))

@contextmanager
def chdir(path):
    from os.path import abspath
    from os import curdir
    path = abspath(path)
    current_dir = abspath(curdir)
    _chdir_and_log(path)
    try:
        yield
    finally:
        _chdir_and_log(current_dir)

@contextmanager
def open_buildout_configfile(filepath="buildout.cfg", write_on_exit=False):
    from ConfigParser import ConfigParser
    parser = ConfigParser()
    parser.read(filepath)
    try:
        yield parser
    finally:
        if not write_on_exit:
            return
        with open(filepath, 'w') as fd:
            parser.write(fd)

def is_running_inside_virtualenv():
    import sys
    return hasattr(sys, 'real_prefix')

def parse_args(commandline_or_args):
    return commandline_or_args if isinstance(commandline_or_args, list) else commandline_or_args.split()

def execute_assert_success(args):
    from infi import execute
    logger.info("Executing {}".format(' '.join(args)))
    result = execute.execute(args)
    if result.get_returncode() is not None and result.get_returncode() != 0:
        raise PrettyExecutionError(result)

def _get_executable_from_shebang_line():  # pragma: no cover
    # The executable wrapper in distribute dynamically loads Python's DLL, which causes sys.executable to be the wrapper
    # and not the original python exeuctable. We have to find the real executable as Distribute does.
    from os import path
    import sys
    executable_script_py = sys.executable.replace(".exe", "-script.py")
    if not path.exists(executable_script_py):
        # using original Python executable
        return sys.executable
    with open(executable_script_py) as fd:
        shebang_line = fd.readlines()[0].strip()
    executable_path = path.normpath(shebang_line[2:])
    return (executable_path + '.exe') if not executable_path.endswith('.exe') else executable_path

def execute_with_python(commandline_or_args):
    import sys
    from ..assertions import is_windows
    args = parse_args(commandline_or_args)
    executable = [sys.executable if not is_windows() else _get_executable_from_shebang_line()]
    if not is_running_inside_virtualenv():
        executable.append('-S')
    try:
        execute_assert_success(executable + args)
    except PrettyExecutionError:
        logger.warning("Command falied with -S, trying without")
        executable.remove('-S')
        execute_assert_success(executable + args)

def execute_with_isolated_python(commandline_or_args):
    import sys
    import os
    from ..assertions import is_windows
    args = parse_args(commandline_or_args)
    executable = [os.path.join('parts', 'python', 'bin', 'python{}'.format('.exe' if is_windows() else ''))]
    with open_buildout_configfile() as buildout:
        if buildout.get('buildout', 'relative-paths') in ['True', 'true']:
            [executable] = os.path.abspath(executable[0])
    execute_assert_success(executable + args)

def execute_with_buildout(commandline_or_args):
    from os import name, path
    args = parse_args(commandline_or_args)
    execute_assert_success([path.join('bin', 'buildout{}'.format('.exe' if name == 'nt' else ''))] + \
                            BUILDOUT_PARAMETERS + args)

@contextmanager
def buildout_parameters_context(parameters):
    try:
        _ = [BUILDOUT_PARAMETERS.append(param) for param in parameters if param not in BUILDOUT_PARAMETERS]
        yield
    finally:
        _ = [BUILDOUT_PARAMETERS.remove(param) for param in parameters if param in BUILDOUT_PARAMETERS]

def _release_version_with_git_flow(version_tag):
    from os import curdir
    from gitflow.core import GitFlow
    from gitpy import LocalRepository
    gitflow = GitFlow()
    gitflow.create("release", version_tag, base=None, fetch=False)
    gitflow.finish("release", version_tag, fetch=False, rebase=False, keep=False, force_delete=True,
                   tagging_info=dict(sign=False, message=version_tag))
    repository = LocalRepository(curdir)
    repository.checkout("develop")
    repository.commit("empty commit after version {}".format(version_tag), allowEmpty=True)
    repository.createTag("{}-develop".format(version_tag))

def git_checkout(branch_name_or_tag):
    from os import curdir
    from gitpy import LocalRepository
    try:
        LocalRepository(curdir).checkout(branch_name_or_tag)
    except Exception: # pragma: no cover
        logger.error("failed to checkout {}".format(branch_name_or_tag))
        raise SystemExit(1)

def commit_changes_to_buildout(message):
    from os import curdir
    from gitpy import LocalRepository
    repository = LocalRepository(curdir)
    if "buildout.cfg" not in [modified_file.filename for modified_file in repository.getChangedFiles()]:
        return
    repository.add("buildout.cfg")
    repository.commit("buildout.cfg: " + message)

def get_latest_version():
    from os import curdir
    from gitpy import LocalRepository
    from pkg_resources import parse_version
    repository = LocalRepository(curdir)
    version_tags = [tag.name for tag in repository.getTags()
                    if tag.name.startswith('v') and not tag.name.endswith('-develop')]
    version_tags.sort(key=lambda ver: parse_version(ver))
    return version_tags[-1]

@contextmanager
def open_tempfile():
    from tempfile import mkstemp
    from os import close, remove
    fd, path = mkstemp()
    close(fd)
    try:
        yield path
    finally:
        try:
            remove(path)
        except:
            pass

def set_freezed_versions_in_install_requires(buildout_cfg, versions_cfg):
    from .package_sets import InstallRequiresPackageSet, VersionSectionSet, from_dict, to_dict
    install_requires = to_dict(InstallRequiresPackageSet.from_value(buildout_cfg.get("project", "install_requires")))
    versions = to_dict(VersionSectionSet.from_value(versions_cfg))
    for key, value in versions.items():
        if install_requires.has_key(key):
            install_requires[key] = value
    install_requires = from_dict(install_requires)
    install_requires = set([item.replace('==', ">=") for item in install_requires])
    buildout_cfg.set("project", "install_requires", InstallRequiresPackageSet.to_value(install_requires))

def freeze_versions(versions_file, change_install_requires):
    from os import curdir, path
    with open_buildout_configfile(write_on_exit=True) as buildout_cfg:
        buildout_cfg.set("buildout", "extensions",  "buildout-versions")
        buildout_cfg.set("buildout", "versions", "versions")
        with open_buildout_configfile(versions_file) as versions_cfg:
            if not buildout_cfg.has_section("versions"):
                buildout_cfg.add_section("versions")
            for option in buildout_cfg.options("versions"):
                buildout_cfg.remove_option("versions", option)
            for option in versions_cfg.options("versions"):
                buildout_cfg.set("versions", option, versions_cfg.get("versions", option))
        if change_install_requires:
                set_freezed_versions_in_install_requires(buildout_cfg, versions_cfg)

def unset_freezed_versions_in_install_requires(buildout_cfg):
    from .package_sets import InstallRequiresPackageSet, to_dict, from_dict
    install_requires = InstallRequiresPackageSet.from_value(buildout_cfg.get("project", "install_requires"))
    install_requires_dict = to_dict(install_requires)
    install_requires_dict.update({key:[] for key, specs in install_requires_dict.items()
                                  if specs and specs[-1][0] == '>='})
    install_requires = from_dict(install_requires_dict)
    buildout_cfg.set("project", "install_requires", InstallRequiresPackageSet.to_value(install_requires))

def unfreeze_versions(change_install_requires):
    with open_buildout_configfile(write_on_exit=True) as buildout_cfg:
        buildout_cfg.remove_option("buildout", "extensions")
        buildout_cfg.remove_option("buildout", "buildout_versions_file")
        buildout_cfg.remove_option("buildout", "extends")
        buildout_cfg.remove_option("buildout", "versions")
        buildout_cfg.remove_section("versions")
        if change_install_requires:
            unset_freezed_versions_in_install_requires(buildout_cfg)

@contextmanager
def revert_if_failed(keep_leftovers):
    from gitpy import LocalRepository
    from os import curdir
    repository = LocalRepository(curdir)
    get_tags = lambda: {tag.name:tag for tag in repository.getTags()}
    get_branches = lambda: {branch.name:branch for branch in repository.getBranches()}
    get_head = lambda branch_name: repository.getBranchByName(branch_name).getHead()
    get_status = lambda: dict(develop=get_head("develop"), master=get_head("master"), tags=get_tags(),
                              branches=get_branches())
    before = get_status()
    try:
        yield
    except:
        if keep_leftovers:
            raise
        now = get_status()
        for tag in set(now['tags']).difference(set(before['tags'])):
            repository.delete(now['tags'][tag])
        for branch in set(now['branches']).difference(set(before['branches'])):
            repository.delete(now['branches'][branch])
        for branch_name in ['master', 'develop']:
            repository.resetHard()
            branch = repository.getBranchByName(branch_name)
            repository.checkout(branch)
            repository.resetHard(before[branch_name])
        raise

def release_version_with_git_flow(version_tag, keep_leftovers=True):
    with revert_if_failed(keep_leftovers):
        _release_version_with_git_flow(version_tag)
